/*========================================================================*\

Copyright (c) 2001-2002  Paul Vojta

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
PAUL VOJTA BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

NOTE:
	xdvi is based on prior work, as noted in the modification history
	in xdvi.c.

\*========================================================================*/

/*
 *	Code for popup windows.
 */

#include "xdvi.h"
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>

/* Xlib and Xutil are already included */

#include <X11/Xatom.h>
#include <X11/StringDefs.h>

#if XAW
# include <X11/Shell.h>
# if XAW3D
#  include <X11/Xaw3d/Form.h>
#  include <X11/Xaw3d/Label.h>
#  include <X11/Xaw3d/Command.h>
#  include <X11/Xaw3d/Toggle.h>
#  include <X11/Xaw3d/AsciiText.h>
# else
#  include <X11/Xaw/Form.h>
#  include <X11/Xaw/Label.h>
#  include <X11/Xaw/Command.h>
#  include <X11/Xaw/Toggle.h>
#  include <X11/Xaw/AsciiText.h>
# endif
# if XtSpecificationRelease < 5
#  define XawChainLeft	XtChainLeft
#  define XawChainRight	XtChainRight
#  define XawChainTop	XtChainTop
#  define XawChainBottom XtChainBottom
# endif
# if XtSpecificationRelease < 6
#  define XawFmt8Bit	FMT8BIT
# endif
#else /* MOTIF */
# include <Xm/BulletinB.h>
# include <Xm/DialogS.h>
# include <Xm/MessageB.h>
# include <Xm/LabelG.h>
# include <Xm/Form.h>
# include <Xm/Frame.h>
# include <Xm/ToggleBG.h>
# include <Xm/Text.h>
# include <Xm/TextF.h>
# include <Xm/PushB.h>
#endif /* MOTIF */

#ifdef X_NOT_STDC_ENV
extern	int	errno;
#endif

#if NeedVarargsPrototypes
# include <stdarg.h>
#else
# include <varargs.h>
#endif

#ifndef va_copy
# ifdef __va_copy
#  define va_copy __va_copy
# else
#  define va_copy(aq, ap) aq = ap
# endif
#endif

#if HAVE_VFORK_H
# include <vfork.h>
#endif

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WIFEXITED
# define WIFEXITED(status)	(((status) & 255) == 0)
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(status)	((unsigned)(status) >> 8)
#endif
#ifndef WIFSIGNALED
# ifndef WIFSTOPPED
#  define WIFSTOPPED(status)	(((status) & 0xff) == 0x7f)
# endif
# define WIFSIGNALED(status)	(!WIFSTOPPED(status) && !WIFEXITED(status))
#endif
#ifndef WTERMSIG
# define WTERMSIG(status)	((status) & 0x7f)
#endif

/* if POSIX O_NONBLOCK is not available, use O_NDELAY */
#if !defined(O_NONBLOCK) && defined(O_NDELAY)
# define O_NONBLOCK O_NDELAY
#endif

#ifdef EWOULDBLOCK
# ifdef EAGAIN
#  define AGAIN_CONDITION	(errno == EWOULDBLOCK || errno == EAGAIN)
# else
#  define AGAIN_CONDITION	(errno == EWOULDBLOCK)
# endif
#else /* EWOULDBLOCK */
# ifdef EAGAIN
#  define AGAIN_CONDITION	(errno == EAGAIN)
# endif
#endif /* EWOULDBLOCK */

#if HAVE_POLL
# include <poll.h>
# define XIO_IN POLLIN
# define XIO_OUT POLLOUT
#else
# define XIO_IN 1
# define XIO_OUT 2
#endif


/*
 *	New variants on printf.
 */

static int uintlen(i)
	unsigned int i;
{
	int	len = 0;

	do {
	    ++len;
	    i /= 10;
	}
	while (i != 0);

	return len;
}

/*
 *	(v)nprintf(fmt, ...) - return the length of the string produced.
 */

#if HAVE_GOOD_VSNPRINTF

# define vnprintf(fmt, ap)	vsnprintf(NULL, 0, fmt, ap)

#else /* !HAVE_GOOD_VSNPRINTF */

static int
vnprintf(format, args)
	_Xconst char	*format;
	va_list		args;
{
	_Xconst char	*p, *p_end, *q;
	int		len;

	p = format;
	p_end = p + strlen(p);
	len = 0;
	for (;;) {
	    q = memchr(p, '%', p_end - p);
	    if (q == NULL) {
		len += p_end - p;
		break;
	    }
	    len += q - p;
	    p = q + 1;
	    switch (*p++) {
		case '%':
		    ++len;
		    break;
		case 'd':
		    {
			int i = va_arg(args, int);

			len += (i < 0 ? uintlen(-i) + 1 : uintlen(i));
		    }
		    break;
		case 's':
		    len += strlen(va_arg(args, _Xconst char *));
		    break;
		default:
		    oops("Invalid designator `%c' in format string \"%s\"\n",
			p[-1], format);
	    }
	}

	return len;
}


#if 0	/* not needed (yet) */

static int
#if NeedVarargsPrototypes
nprintf(_Xconst char *format, ...)
#else
/* VARARGS */
nprintf(va_alist)
	va_dcl
#endif
{
#if !NeedVarargsPrototypes
	_Xconst char	*format;
#endif
	va_list		args;
	int		len;

#if NeedVarargsPrototypes
	va_start(args, format);
#else
	va_start(args);
	format = va_arg(args, _Xconst char *);
#endif

	len = vnprintf(format, args);
	va_end(args);
	return len;
}

#endif /* if 0 */

#endif /* !HAVE_GOOD_VSNPRINTF */

/*
 *	(v)mprintf(fmt, ...) - printf to malloc()ed string.
 */

static char *
vmprintf(format, args)
	_Xconst char	*format;
	va_list		args;
{
	va_list		args2;
	int		len;
	char		*result;

	va_copy(args2, args);
	len = vnprintf(format, args2) + 1;
	va_end(args2);

	result = xmalloc(len);
#if HAVE_VSNPRINTF
	(void) vsnprintf(result, len, format, args);
#else
	(void) vsprintf(result, format, args);
#endif

	return result;
}

static char *
#if NeedVarargsPrototypes
mprintf(_Xconst char *format, ...)
#else
/* VARARGS */
mprintf(va_alist)
	va_dcl
#endif
{
#if !NeedVarargsPrototypes
	_Xconst char	*format;
#endif
	va_list		args;
	char		*result;

#if NeedVarargsPrototypes
	va_start(args, format);
#else
	va_start(args);
	format = va_arg(args, _Xconst char *);
#endif

	result = vmprintf(format, args);
	va_end(args);
	return result;
}


/*
 *	Simple popup - Just a text message.  May not be done before realizing
 *	the main window.
 */

/*
 *	In reality, the "simple" popup is not so simple, because certain
 *	window managers (e.g., twm, fvwm1) leave the window showing if you
 *	pop it down too soon after popping it up.  To remedy this, we delay
 *	calling XtPopdown() until the previous call to XtPopup() has generated
 *	a MapNotify event.  Likewise, we delay XtPopup() until the previous
 *	XtPopdown() call has produced an UnmapNotify event.
 *
 *	To make it just a little more complicated, for some reason the popup
 *	may get mapped after being unmapped.  If so, we call XtPopup() to
 *	inform the toolkit of its mistake, and arrange to expect some spurious
 *	events.  To reproduce this bug, remove the "spurious" code, run xdvi,
 *	move the mouse a little, then hit "n" a few times.  This happens only
 *	with twm.
 */

/* ARGSUSED */
static	void
handle_popup_events(widget, rec, ev, cont)
	Widget	widget;
	XtPointer rec;
	XEvent	*ev;
	Boolean	*cont;		/* unused */
{
#define	prec	((struct status_popup *) rec)

	if (ev->type == MapNotify) {
	    if (prec->expected_type == MapNotify) {
		prec->expected_type = 0;
		if (!prec->popped) {
		    prec->expected_type = UnmapNotify;
		    XtPopdown(prec->shell);
		}
	    }
	    else {
		if (debug & DBG_EVENT)
		    fprintf(stderr,
		      "handle_popup_events: got spurious MapNotify\n");
		prec->expected_type = UnmapNotify;
		prec->spurious = True;
		XtPopup(prec->shell, XtGrabNone);
	    }
	}
	else if (ev->type == UnmapNotify) {
	    if (prec->expected_type == UnmapNotify) {
		prec->expected_type = 0;
		if (prec->spurious) {
		    prec->expected_type = MapNotify;
		    prec->spurious = False;
		}
		else if (prec->popped) {
		    prec->expected_type = MapNotify;
		    XtPopup(prec->shell, XtGrabNone);
		}
	    }
	    else
		if (debug & DBG_EVENT)
		    fprintf(stderr,
		      "handle_popup_events: got spurious UnmapNotify\n");
	}

#undef	prec
}


#if XAW

void
simple_popup(rec, message)
	struct status_popup	*rec;
	_Xconst char		*message;
{
	Position	x, y;
	Dimension	w1, h1, w2, h2;
	Boolean		old_rec_popped	= rec->popped;

	rec->popped = True;
	if (rec->shell == NULL) {
	    rec->shell = XtVaCreatePopupShell("status",
	      transientShellWidgetClass, top_level,
	      XtNtitle, "Xdvi status",
	      XtNallowShellResize, True,
	      XtNtransientFor, top_level,
	      XtNvisual, our_visual,
	      XtNcolormap, our_colormap,
	      NULL);
	    XtAddEventHandler(rec->shell, StructureNotifyMask, False,
	      handle_popup_events, (XtPointer) rec);

	    rec->label = XtVaCreateManagedWidget("label", labelWidgetClass,
	      rec->shell,
	      XtNlabel, message,
	      NULL);
	}
	else
	    XtVaSetValues(rec->label, XtNlabel, message, NULL);

	if (XtIsRealized(top_level))
	    XtVaGetValues(top_level, XtNx, &x, XtNy, &y,
	      XtNwidth, &w1, XtNheight, &h1, NULL);
	else {
	    x = y = 0;
	    w1 = WidthOfScreen(SCRN);
	    h1 = HeightOfScreen(SCRN);
	}

	/* Get the size of the popup window */
	XtVaGetValues(rec->label, XtNwidth, &w2, XtNheight, &h2, NULL);

	/* Center the popup over the main window */
	XtVaSetValues(rec->shell, XtNx, x + (w1 - w2) / 2,
	  XtNy, y + (h1 - h2) / 2, NULL);

	if (!rec->expected_type && !old_rec_popped) {
	    rec->expected_type = MapNotify;
	    XtPopup(rec->shell, XtGrabNone);
	}
}

#else /* MOTIF */

void
simple_popup(rec, message)
	struct status_popup	*rec;
	_Xconst char		*message;
{
	Widget		bb;
	XmString	str;
	Boolean		old_rec_popped	= rec->popped;

	rec->popped = True;
	str = XmStringCreateLtoR((char *) message, XmFONTLIST_DEFAULT_TAG);
	if (rec->shell == NULL) {
	    rec->shell = XtVaCreatePopupShell("status",
	      xmDialogShellWidgetClass, top_level,
	      XmNallowShellResize, True,
	      XmNvisual, our_visual,
	      XmNcolormap, our_colormap,
	      XmNtitle, "Xdvi status",
	      NULL);
	    XtAddEventHandler(rec->shell, StructureNotifyMask, False,
	      handle_popup_events, (XtPointer) rec);

	    bb = XtCreateWidget("bulletin", xmBulletinBoardWidgetClass,
	      rec->shell, NULL, 0);

	    rec->label = XtVaCreateManagedWidget("label", xmLabelGadgetClass,
	      bb,
	      XmNlabelString, str,
	      NULL);

	    XtManageChild(bb);
	}
	else
	    XtVaSetValues(rec->label, XmNlabelString, str, NULL);

	XmStringFree(str);

	if (!rec->expected_type && !old_rec_popped) {
	    rec->expected_type = MapNotify;
	    XtPopup(rec->shell, XtGrabNone);
	}
}

#endif

void
simple_popdown(rec)
	struct status_popup	*rec;
{
	rec->popped = False;
	if (!rec->expected_type) {
	    rec->expected_type = UnmapNotify;
	    XtPopdown(rec->shell);
	}
}


/*
 *	Warning popup - used for most error and notice conditions.
 *	It includes a message and a button (and, in the case of the Motif
 *	toolkit, a bitmap).
 */

#if XAW

/* ARGSUSED */
static	void
warn_callback(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	XtDestroyWidget((Widget) client_data);
}

void
do_popup(shell)
	Widget		shell;
{
	Position	x, y;
	Dimension	w1, h1, w2, h2;

	XtVaGetValues(top_level, XtNx, &x, XtNy, &y,
	  XtNwidth, &w1, XtNheight, &h1, NULL);

	/* Get the size of the popup window */
	XtRealizeWidget(shell);
	XtVaGetValues(shell, XtNwidth, &w2, XtNheight, &h2, NULL);

	/* Center the popup over the main window */
	XtVaSetValues(shell, XtNx, x + (w1 - w2) / 2,
	  XtNy, y + (h1 - h2) / 2, NULL);

	XtPopup(shell, XtGrabNone);
}

Widget
warning_popup(message, button_name, callback)
	_Xconst char	*message;
	_Xconst char	*button_name;
	XtCallbackProc	callback;
{
	Widget		shell, form, label, button;
	Dimension	w1, w2, bw;
	int		dist;

	shell = XtVaCreatePopupShell("warn", transientShellWidgetClass,
	  top_level,
	  XtNtitle, "Xdvi warning",
	  XtNmappedWhenManaged, False,
	  XtNtransientFor, top_level,
	  XtNvisual, our_visual,
	  XtNcolormap, our_colormap,
	  NULL);
	form = XtCreateManagedWidget("form", formWidgetClass, shell,
	  NULL, 0);
	label = XtVaCreateManagedWidget("label", labelWidgetClass, form,
	  XtNlabel, message,
	  XtNborderWidth, 0,
	  XtNleft, XtChainLeft,
	  XtNright, XtChainLeft,
	  NULL);
	button = XtVaCreateManagedWidget(button_name, commandWidgetClass, form,
	  XtNaccelerators, accels_cr_click,
	  XtNfromVert, label,
	  NULL);
	XtAddCallback(button, XtNcallback,
	  callback != NULL ? callback : warn_callback, shell);
	XtInstallAccelerators(form, button);

	/* Center button under message */
	XtVaGetValues(label, XtNwidth, &w1, NULL);
	XtVaGetValues(button, XtNwidth, &w2, XtNborderWidth, &bw, NULL);
	w2 += 2 * bw;
	if (w1 > w2) {
	    XtVaGetValues(button, XtNhorizDistance, &dist, NULL);
	    XtVaSetValues(button, XtNhorizDistance, dist + (w1 - w2) / 2, NULL);
	}
	else if (w1 < w2) {
	    XtVaGetValues(label, XtNhorizDistance, &dist, NULL);
	    XtVaSetValues(label, XtNhorizDistance, dist + (w2 - w1) / 2, NULL);
	}

	if (!XtIsRealized(top_level)) {
	    /* save it for later */
	    if (n_init_popups >= alloc_init_popups) {
		if (alloc_init_popups != 0)
		    init_popups = xrealloc(init_popups,
		      (alloc_init_popups += 2) * sizeof (*init_popups));
		else
		    init_popups = xmalloc((alloc_init_popups = 2)
		      * sizeof (*init_popups));
	    }
	    init_popups[n_init_popups++] = shell;
	}
	else
	    do_popup(shell);

	return shell;
}

#else /* MOTIF */

/* ARGSUSED */
static	void
warn_callback(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	XtDestroyWidget((Widget) client_data);
}

void
do_popup(dialog)
	Widget		dialog;
{
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
}

Widget
warning_popup(message, mtype, button_name, callback)
	_Xconst char	*message;
	int		mtype;
	_Xconst char	*button_name;
	XtCallbackProc	callback;
{
	Widget		shell;
	Widget		dialog;
	XmString	str;

	shell = XtVaCreatePopupShell("warn", xmDialogShellWidgetClass,
	  top_level,
	  XmNvisual, our_visual,
	  XmNcolormap, our_colormap,
	  XmNtitle, "Xdvi warning",
	  NULL);
	dialog = XtVaCreateWidget("dialog", xmMessageBoxWidgetClass, shell,
	  XmNdialogType, mtype,
	  NULL);
	XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
	XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));

	XtAddCallback(dialog, XmNokCallback,
	  callback != NULL ? callback : warn_callback, shell);

	str = XmStringCreateLtoR((char *) message, XmFONTLIST_DEFAULT_TAG);
	XtVaSetValues(dialog, XmNmessageString, str, NULL);
	XmStringFree(str);

	if (button_name != NULL) {
	    str = XmStringCreateLocalized((char *) button_name);
	    XtVaSetValues(dialog, XmNokLabelString, str, NULL);
	    XmStringFree(str);
	}

	XtInstallAllAccelerators(dialog,
	  XmMessageBoxGetChild(dialog, XmDIALOG_OK_BUTTON));

	if (!XtIsRealized(top_level)) {
	    /* save it for later */
	    if (n_init_popups >= alloc_init_popups) {
		if (alloc_init_popups != 0)
		    init_popups = xrealloc(init_popups,
		      (alloc_init_popups += 2) * sizeof (*init_popups));
		else
		    init_popups = xmalloc((alloc_init_popups = 2)
		      * sizeof (*init_popups));
	    }
	    init_popups[n_init_popups++] = dialog;
	}
	else
	    do_popup(dialog);

	return shell;
}

#endif /* MOTIF */

Widget
#if NeedVarargsPrototypes
warning_popup_long(_Xconst char *format,
# if MOTIF
  int mtype,
# endif
  _Xconst char *button_name, XtCallbackProc callback, ...)
#else /* ! NeedVarargsPrototypes */
# if XAW
warning_popup_long(format, button_name, callback, va_alist)
	_Xconst char	*format;
	_Xconst char	*button_name;
	XtCallbackProc	callback;
# else /* MOTIF */
warning_popup_long(format, mtype, button_name, callback, va_alist)
	_Xconst char	*format;
	int		mtype;
	_Xconst char	*button_name;
	XtCallbackProc	callback;
# endif
	va_dcl
#endif /* ! NeedVarargsPrototypes */
{
	va_list		args;
	char		*str;
	Widget		w;

#if NeedVarargsPrototypes
	va_start(args, callback);
#else
	va_start(args);
#endif
	str = vmprintf(format, args);
	va_end(args);

#if XAW
	w = warning_popup(str, button_name, callback);
#else /* MOTIF */
	w = warning_popup(str, mtype, button_name, callback);
#endif

	free(str);

	return w;
}


#if XAW

/*
 *	Confirm popup - Pop up a window and ask a yes/no question.
 */

Widget
confirm_popup(message, button_name, callback, yes_answer, no_answer)
	_Xconst char	*message;
	_Xconst char	*button_name;
	XtCallbackProc	callback;
	XtPointer	yes_answer, no_answer;
{
	Widget		shell, form, label, button1, button2;
	int		ddist;
	Dimension	w0, w1, w2, bw1, bw2, tw;

	shell = XtVaCreatePopupShell("confirm", transientShellWidgetClass,
	  top_level,
	  XtNtitle, "Xdvi question",
	  XtNmappedWhenManaged, False,
	  XtNtransientFor, top_level,
	  XtNvisual, our_visual,
	  XtNcolormap, our_colormap,
	  NULL);
	form = XtCreateManagedWidget("form", formWidgetClass, shell,
	  NULL, 0);
	label = XtVaCreateManagedWidget("label", labelWidgetClass, form,
	  XtNlabel, message,
	  XtNborderWidth, 0,
	  XtNleft, XtChainLeft,
	  XtNright, XtChainLeft,
	  NULL);

	button1 = XtVaCreateManagedWidget(button_name, commandWidgetClass, form,
	  XtNaccelerators, accels_cr,
	  XtNfromVert, label,
	  NULL);
	XtAddCallback(button1, XtNcallback, callback, yes_answer);
	XtInstallAccelerators(form, button1);

	button2 = XtVaCreateManagedWidget("Cancel", commandWidgetClass, form,
	  XtNfromVert, label,
	  XtNfromHoriz, button1,
	  NULL);
	XtAddCallback(button2, XtNcallback, callback, no_answer);

	/* Move cancel button to the right */
	XtVaGetValues(label, XtNwidth, &w0, NULL);
	XtVaGetValues(button1, XtNwidth, &w1, XtNborderWidth, &bw1, NULL);
	XtVaGetValues(button2, XtNwidth, &w2, XtNborderWidth, &bw2, NULL);
	XtVaGetValues(form, XtNdefaultDistance, &ddist, NULL);
	tw = w1 + w2 + 2 * (bw1 + bw2);
	if (tw < w0 + ddist)
	    XtVaSetValues(button2, XtNhorizDistance, w0 - tw, NULL);

	do_popup(shell);

	return shell;
}

#else /* MOTIF */

Widget
confirm_popup(message, button_name, callback, yes_answer, no_answer)
	char		*message;
	char		*button_name;
	XtCallbackProc	callback;
	XtPointer	yes_answer, no_answer;
{
	Widget		dialog;
	XmString	msg, yes, no;

	{
	    static Arg	cq_args[] = {
		{XmNvisual,	(XtArgVal) 0},
		{XmNcolormap,	(XtArgVal) 0},
	    };

	    cq_args[0].value = (XtArgVal) our_visual;
	    cq_args[1].value = (XtArgVal) our_colormap;
	    dialog = XmCreateQuestionDialog(top_level, "confirm",
	      cq_args, XtNumber(cq_args));
	}
	msg = XmStringCreateLocalized(message);
	yes = XmStringCreateLocalized(button_name);
	no = XmStringCreateLocalized("Cancel");
	XtVaSetValues(dialog,
	  XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
	  XmNmessageString, msg,
	  XmNokLabelString, yes,
	  XmNcancelLabelString, no,
	  NULL);
	XmStringFree(msg);
	XmStringFree(yes);
	XmStringFree(no);

	XtAddCallback(dialog, XmNokCallback, callback, yes_answer);
	XtAddCallback(dialog, XmNcancelCallback, callback, no_answer);
	XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));

	XtManageChild(dialog);
	dialog = XtParent(dialog);
	XtPopup(dialog, XtGrabNone);

	return dialog;
}

#endif /* MOTIF */


/*
 *	Print popup - pop up a window to allow the user to print page(s)
 *	from the dvi file.
 */

static	Boolean	print_active	= False;	/* if print window is showing */
static	Widget	print_shell	= NULL;		/* shell widget of popup*/
static	Widget	r2label, r2text;
static	Widget	r3label, r3text;
static	Widget	r4text;
static	Widget	r7text1, r7text2;
static	Widget	r8text;
static	Widget	r9prev;
static	int	pr_current_fileradio	= 1;	/* 1 if printer, 2 if file */
static	int	pr_current_pageradio	= 1;	/* 1 = whole file, 2 = range */

/* Stuff to save between popups */

static	int		pr_save_fileradio = 1;
static	int		pr_save_pageradio = 1;
static	_Xconst char	*pr_save_cmd	= NULL;
static	_Xconst char	*pr_save_file	= NULL;
static	_Xconst char	*pr_save_printer= NULL;
static	_Xconst char	*pr_save_xargs	= NULL;

/* Things to do with running the dvips process */

static	Boolean	printlog_active	= False;	/* if the print log is active */
static	Widget	printlog_shell	= NULL;
static	Widget	printlog_keep;

/* Forward reference */

static	void	print_precheck();


static void
set_sensitivity(i)
	int	i;
{
	Boolean	sensitivity;

	if (i == pr_current_fileradio) /* if no change */
	    return;

	pr_current_fileradio = i;

	sensitivity = (i == 1);
	XtSetSensitive(r2label, sensitivity);
	XtSetSensitive(r2text, sensitivity);

	sensitivity ^= (True ^ False);
	XtSetSensitive(r3label, sensitivity);
	XtSetSensitive(r3text, sensitivity);
}

/* ARGSUSED */
static	void
cb_print_cancel(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	XtPopdown(print_shell);
	print_active = False;

	if (w == r9prev && !printlog_active) {
	    XtSetSensitive(printlog_keep, False);
	    XtPopup(printlog_shell, XtGrabNone);
	    printlog_active = True;
	}
}

/* ARGSUSED */
static	void
cb_print(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	print_precheck();
}

#if XAW
static void print_act_go ARGS((Widget, XEvent *, String *, Cardinal *));
#endif
static void printlog_act_close ARGS((Widget, XEvent *, String *, Cardinal *));
static void printlog_act_keep ARGS((Widget, XEvent *, String *, Cardinal *));
static void printlog_act_unkeep ARGS((Widget, XEvent *, String *, Cardinal *));
static void printlog_act_cancel ARGS((Widget, XEvent *, String *, Cardinal *));

static	XtActionsRec	print_actions[]	= {
#if XAW
	{"printInternal",	print_act_go},
#endif
	{"printlogIntClose",	printlog_act_close},
	{"printlogIntKeep",	printlog_act_keep},
	{"printlogIntUnkeep",	printlog_act_unkeep},
	{"printlogIntCancel",	printlog_act_cancel},
};

#if XAW

#ifndef PRINTINDENT
# define PRINTINDENT	25
#endif

static	Widget	r1radio1;
static	Widget	r6radio, r7radio;

/* ARGSUSED */
static	void
cb_print_vs_file(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	int	i	= (int) XawToggleGetCurrent(r1radio1);

	if (i != 0)
	    set_sensitivity(i);
}

/* ARGSUSED */
static	void
cb_range(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	int	i	= (int) XawToggleGetCurrent(r6radio);

	if (i != 0)
	    pr_current_pageradio = i;
}

/* ARGSUSED */
void
range_handle_key(widget, closure, ev, cont)
	Widget		widget;
	XtPointer	closure;
	XEvent		*ev;
	Boolean		*cont;
{
	if (pr_current_pageradio != 2)
	    XawToggleSetCurrent(r6radio, (XtPointer) 2);
}

/* ARGSUSED */
static	void
print_act_go P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
	print_precheck();
}

/* ARGSUSED */
void
Act_print P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
	static	Widget	r7label3;
	static	Widget	r9cancel;
	static	Dimension sw;	/* shell width */
	static	int	canceladjust;
	char		*ofstring;

	if (print_active) {
	    XRaiseWindow(DISP, XtWindow(print_shell));
	    return;
	}

	if (printlog_active) {
	    XRaiseWindow(DISP, XtWindow(printlog_shell));
	    return;
	}

	if (dvi_file == NULL) {
	    WARN(XmDIALOG_ERROR, "No active dvi file");
	    return;
	}

	ofstring = (pageno_correct == 1 ? mprintf("of %d", total_pages)
	  : mprintf("of %d to %d", pageno_correct,
	    total_pages + pageno_correct - 1));

	if (print_shell == NULL) {
	    Widget		form;
	    Widget		r1label, r1radio2;
	    Widget		r4label;
	    Widget		r5label;
	    Widget		r6form;
	    Widget		r7form, r7label1, r7label2;
	    Widget		r8label;
	    Widget		r9ok;
	    XtTranslations	xlats, xlats2;
	    XtAccelerators	accels2;
	    int			ddist;

	    XtAddActions(print_actions, XtNumber(print_actions));
	    print_shell = XtVaCreatePopupShell("print",
	      transientShellWidgetClass,
	      top_level,
	      XtNtitle, "Print dvi file",
	      XtNmappedWhenManaged, False,
	      XtNtransientFor, top_level,
	      XtNallowShellResize, True,
	      XtNvisual, our_visual,
	      XtNcolormap, our_colormap,
	      NULL);

	    form = XtCreateManagedWidget("form", formWidgetClass, print_shell,
	      NULL, 0);
	    XtVaGetValues(form, XtNdefaultDistance, &ddist, NULL);

	    r1label = XtVaCreateManagedWidget("printto", labelWidgetClass, form,
	      XtNlabel, "Print to:",
	      XtNborderWidth, 0,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    xlats = XtParseTranslationTable("<EnterWindow>:highlight(Always)\n\
<LeaveWindow>:unhighlight()\n\
<Btn1Down>,<Btn1Up>:set()notify()");
	    r1radio1 = XtVaCreateManagedWidget("toprinter", toggleWidgetClass,
	      form,
	      XtNlabel, "Printer",
	      XtNradioData, 1,
	      XtNstate, True,
	      XtNtranslations, xlats,
	      XtNfromHoriz, r1label,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtAddCallback(r1radio1, XtNcallback, cb_print_vs_file, NULL);

	    r1radio2 = XtVaCreateManagedWidget("tofile", toggleWidgetClass,
	      form,
	      XtNlabel, "File",
	      XtNradioGroup, r1radio1,
	      XtNradioData, 2,
	      XtNtranslations, xlats,
	      XtNfromHoriz, r1radio1,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtAddCallback(r1radio2, XtNcallback, cb_print_vs_file, NULL);

	    /* Rows 2, 3:  Print command or file name */

	    r2label = XtVaCreateManagedWidget("prncommand", labelWidgetClass,
	      form,
	      XtNlabel, "Print command:",
	      XtNborderWidth, 0,
	      XtNfromVert, r1radio1,
	      XtNhorizDistance, ddist + PRINTINDENT,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r2text = XtVaCreateManagedWidget("prntext", asciiTextWidgetClass,
	      form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNfromHoriz, r2label,
	      XtNfromVert, r1radio1,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    xlats2 = XtParseTranslationTable("<Key>Return:printInternal()");
	    XtOverrideTranslations(r2text, xlats2);

	    r3label = XtVaCreateManagedWidget("filename", labelWidgetClass,
	      form,
	      XtNlabel, "File name:",
	      XtNsensitive, False,
	      XtNborderWidth, 0,
	      XtNfromVert, r2text,
	      XtNhorizDistance, ddist + PRINTINDENT,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r3text = XtVaCreateManagedWidget("filetext", asciiTextWidgetClass,
	      form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNsensitive, False,
	      XtNfromHoriz, r3label,
	      XtNfromVert, r2text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtOverrideTranslations(r3text, xlats2);

	    {	/* align left edges of asciitext widgets */
		Dimension w2, w3;

		XtVaGetValues(r2label, XtNwidth, &w2, NULL);
		XtVaGetValues(r3label, XtNwidth, &w3, NULL);
		if (w3 > w2)
		    XtVaSetValues(r2text, XtNhorizDistance, ddist + (w3 - w2),
		      NULL);
		else if (w2 > w3)
		    XtVaSetValues(r3text, XtNhorizDistance, ddist + (w2 - w3),
		      NULL);
	    }

	    /* Row 4:  printer name (for dvips) */

	    r4label = XtVaCreateManagedWidget("prname", labelWidgetClass,
	      form,
	      XtNlabel, "Printer name (used by dvips):",
	      XtNborderWidth, 0,
	      XtNfromVert, r3text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r4text = XtVaCreateManagedWidget("prtext", asciiTextWidgetClass,
	      form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNwidth, 50,
	      XtNfromHoriz, r4label,
	      XtNfromVert, r3text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtOverrideTranslations(r4text, xlats2);

	    /* Rows 5-7:  page selection */

	    r5label = XtVaCreateManagedWidget("rangelab", labelWidgetClass,
	      form,
	      XtNlabel, "Print range:",
	      XtNborderWidth, 0,
	      XtNfromVert, r4text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r6form = XtVaCreateManagedWidget("rangeallform", formWidgetClass,
	      form,
	      XtNdefaultDistance, 0,
	      XtNborderWidth, 0,
	      XtNfromVert, r5label,
	      XtNhorizDistance, ddist + PRINTINDENT,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    accels2 = XtParseAcceleratorTable(
	      "<Btn1Down>,<Btn1Up>:set()notify()");
	    r6radio = XtVaCreateManagedWidget("rangeall", toggleWidgetClass,
	      r6form,
	      XtNlabel, " ",
	      XtNradioData, 1,
	      XtNstate, True,
	      XtNtranslations, xlats,
	      XtNaccelerators, accels2,
	      NULL);
	    XtAddCallback(r6radio, XtNcallback, cb_range, NULL);
	    XtInstallAccelerators(r6form, r6radio);

	    (void) XtVaCreateManagedWidget("rangealllab", labelWidgetClass,
	      r6form,
	      XtNlabel, "All",
	      XtNborderWidth, 0,
	      XtNfromHoriz, r6radio,
	      XtNhorizDistance, ddist,
	      NULL);

	    r7form = XtVaCreateManagedWidget("rangefromtoform", formWidgetClass,
	      form,
	      XtNdefaultDistance, 0,
	      XtNborderWidth, 0,
	      XtNfromVert, r6form,
	      XtNhorizDistance, ddist + PRINTINDENT,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r7radio = XtVaCreateManagedWidget("rangefromto", toggleWidgetClass,
	      r7form,
	      XtNlabel, " ",
	      XtNradioGroup, r6radio,
	      XtNradioData, 2,
	      XtNtranslations, xlats,
	      XtNaccelerators, accels2,
	      NULL);
	    XtAddCallback(r7radio, XtNcallback, cb_range, NULL);
	    XtInstallAccelerators(r7form, r7radio);

	    r7label1 = XtVaCreateManagedWidget("rangefromlab", labelWidgetClass,
	      r7form,
	      XtNlabel, "From",
	      XtNborderWidth, 0,
	      XtNfromHoriz, r7radio,
	      XtNhorizDistance, ddist,
	      NULL);

	    r7text1 = XtVaCreateManagedWidget("rangefrom", asciiTextWidgetClass,
	      r7form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNwidth, 50,
	      XtNfromHoriz, r7label1,
	      XtNhorizDistance, ddist,
	      NULL);
	    XtOverrideTranslations(r7text1, xlats2);
	    XtAddEventHandler(r7text1, KeyPressMask, False,
	      range_handle_key, (XtPointer) NULL);

	    r7label2 = XtVaCreateManagedWidget("rangetolab", labelWidgetClass,
	      r7form,
	      XtNlabel, "to",
	      XtNborderWidth, 0,
	      XtNfromHoriz, r7text1,
	      XtNhorizDistance, ddist,
	      NULL);

	    r7text2 = XtVaCreateManagedWidget("rangeto", asciiTextWidgetClass,
	      r7form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNwidth, 50,
	      XtNfromHoriz, r7label2,
	      XtNhorizDistance, ddist,
	      NULL);
	    XtOverrideTranslations(r7text2, xlats2);
	    XtAddEventHandler(r7text2, KeyPressMask, False,
	      range_handle_key, (XtPointer) NULL);

	    r7label3 = XtVaCreateManagedWidget("rangeof", labelWidgetClass,
	      r7form,
	      XtNlabel, ofstring,
	      XtNborderWidth, 0,
	      XtNresizable, True,
	      XtNfromHoriz, r7text2,
	      XtNhorizDistance, ddist,
	      NULL);

	    /* Row 8:  additional dvips args */

	    r8label = XtVaCreateManagedWidget("xargsname", labelWidgetClass,
	      form,
	      XtNlabel, "Additional dvips arguments (optional):",
	      XtNborderWidth, 0,
	      XtNfromVert, r7form,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);

	    r8text = XtVaCreateManagedWidget("xargstext", asciiTextWidgetClass,
	      form,
	      XtNdataCompression, False,
	      XtNeditType, XawtextEdit,
	      XtNwidth, 50,
	      XtNresizable, True,
	      XtNfromVert, r8label,
	      XtNhorizDistance, ddist + PRINTINDENT,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainRight,
	      NULL);
	    XtOverrideTranslations(r8text, xlats2);

	    /* Row 9:  action buttons */

	    r9ok = XtVaCreateManagedWidget("ok", commandWidgetClass, form,
	      XtNlabel, "OK",
	      XtNaccelerators, accels_cr,
	      XtNfromVert, r8text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtAddCallback(r9ok, XtNcallback, cb_print, NULL);
	    XtInstallAccelerators(form, r9ok);
	    XtInstallAccelerators(r2text, r9ok);
	    XtInstallAccelerators(r3text, r9ok);

	    r9prev = XtVaCreateManagedWidget("prev", commandWidgetClass, form,
	      XtNlabel, "Show previous log",
	      XtNsensitive, False,
	      XtNfromHoriz, r9ok,
	      XtNfromVert, r8text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      NULL);
	    XtAddCallback(r9prev, XtNcallback, cb_print_cancel, NULL);

	    r9cancel = XtVaCreateManagedWidget("cancel", commandWidgetClass,
	      form,
	      XtNlabel, "Cancel",
	      XtNfromHoriz, r9prev,
	      XtNfromVert, r8text,
	      XtNleft, XawChainRight,
	      XtNright, XawChainRight,
	      NULL);
	    XtAddCallback(r9cancel, XtNcallback, cb_print_cancel, NULL);

	    {	/* Center the popup over the main window */
		Dimension	sh;	/* sw was defined earlier */
		Position	tx, ty;
		Dimension	tw, th;

		/* Get the size of the popup window */
		XtRealizeWidget(print_shell);
		XtVaGetValues(print_shell, XtNwidth, &sw, XtNheight, &sh, NULL);

		/* Center the popup over the main window */
		XtVaGetValues(top_level, XtNx, &tx, XtNy, &ty,
		  XtNwidth, &tw, XtNheight, &th, NULL);
		XtVaSetValues(print_shell, XtNx, tx + (tw - sw) / 2,
		  XtNy, ty + (th - sh) / 2, NULL);
	    }

	    {
		/* Set width of dvips-xargs text widget */
		Dimension	b;

		XtVaGetValues(r8text, XtNborderWidth, &b, NULL);
		XtVaSetValues(r8text,
		  XtNwidth, sw - 2 * (b + ddist) - PRINTINDENT,
		  NULL);
	    }

	    {	/* Set canceladjust */
		Dimension	cw, cb;
		Position	cx;

		XtVaGetValues(r9cancel, XtNwidth, &cw, XtNborderWidth, &cb,
		  XtNx, &cx, NULL);
		canceladjust = cx + cw + 2 * cb - ddist;
	    }
	}
	else {	/* if the window was already created */
	    XawToggleSetCurrent(r1radio1, (XtPointer) pr_save_fileradio);
	    set_sensitivity(pr_save_fileradio);
	    XawToggleSetCurrent(r6radio, (XtPointer) pr_save_pageradio);

	    XtVaSetValues(r7label3, XtNlabel, ofstring, NULL);

	    if (pr_save_printer != NULL) {
		XtVaSetValues(r4text, XtNstring, pr_save_printer, NULL);
		XawTextSetInsertionPoint(r4text, strlen(pr_save_printer));
	    }
	    else
		XtVaSetValues(r4text, XtNstring, "", NULL);

	    if (pr_save_xargs != NULL) {
		XtVaSetValues(r8text, XtNstring, pr_save_xargs, NULL);
		XawTextSetInsertionPoint(r8text, strlen(pr_save_xargs));
	    }
	    else
		XtVaSetValues(r8text, XtNstring, "", NULL);

	    XtVaGetValues(print_shell, XtNwidth, &sw, NULL);
	}

	free(ofstring);
	XtVaSetValues(r9prev, XtNhorizDistance, (sw - canceladjust) / 2, NULL);
	XtVaSetValues(r9cancel, XtNhorizDistance, (sw - canceladjust) / 2,
	  NULL);

	{	/* set initial values of print command and file name fields */
	    _Xconst char *s;
	    char s2[3 * sizeof(int) + 1];

	    s = pr_save_cmd;
	    if (s == NULL) s = "lpr";
	    /* for some reason two separate XtVaSetValues calls are necessary */
	    XtVaSetValues(r2text, XtNstring, s, NULL);
	    XawTextSetInsertionPoint(r2text, strlen(s));

	    if (pr_save_file == NULL) {
		char		*s2;
		int		n;

		s = rindex(dvi_name, '/');
		if (s == NULL) s = dvi_name;
		else ++s;
		n = strlen(s);
		if (n >= 4 && memicmp(s + n - 4, ".dvi", 4) == 0) n -= 4;
		s2 = xmalloc(n + 4);
		memcpy(s2, s, n);
		memcpy(s2 + n, ".ps", 4);
		pr_save_file = s2;
	    }
	    s = pr_save_file;
	    XtVaSetValues(r3text, XtNstring, s, NULL);
	    XawTextSetInsertionPoint(r3text, strlen(s));

#if HAVE_VSNPRINTF
	    (void) snprintf(s2, sizeof s2, "%d", current_page + pageno_correct);
#else
	    (void) sprintf(s2, "%d", current_page + pageno_correct);
#endif
	    XtVaSetValues(r7text1, XtNstring, s2, NULL);
	    XawTextSetInsertionPoint(r7text1, strlen(s2));
	    XtVaSetValues(r7text2, XtNstring, s2, NULL);
	    XawTextSetInsertionPoint(r7text2, strlen(s2));
	}

	XtPopup(print_shell, XtGrabNone);
	print_active = True;
}

#else /* MOTIF */

#ifndef DDIST
# define DDIST		4
#endif

#ifndef DDIST_MAJOR
# define DDIST_MAJOR	10
#endif

static	Widget	f1radiowidget;
static	Widget	f2radiowidget;
static	Widget	r7label1, r7label2, r7label3;

/* ARGSUSED */
static	void
cb_print_vs_file(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	if (((XmToggleButtonCallbackStruct *) call_data)->set) {
	    if (w != f1radiowidget) {
		XmToggleButtonGadgetSetState(f1radiowidget, False, False);
		f1radiowidget = w;
		set_sensitivity((int) client_data);
		XmProcessTraversal(pr_current_fileradio == 1 ? r2text : r3text,
		  XmTRAVERSE_CURRENT);
	    }
	}
	else if (w == f1radiowidget)
	    XmToggleButtonGadgetSetState(w, True, False);
}


static void
range_set_sensitivity(i)
	int	i;
{
	Boolean	sensitivity;

	if (i == pr_current_pageradio) /* if no change */
	    return;

	pr_current_pageradio = i;
	sensitivity = (i == 2);

	XtSetSensitive(r7label1, sensitivity);
	XtSetSensitive(r7text1, sensitivity);
	XtSetSensitive(r7label2, sensitivity);
	XtSetSensitive(r7text2, sensitivity);
	XtSetSensitive(r7label3, sensitivity);
}

/* ARGSUSED */
static	void
cb_range(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	if (((XmToggleButtonCallbackStruct *) call_data)->set) {
	    if (w != f2radiowidget) {
		XmToggleButtonGadgetSetState(f2radiowidget, False, False);
		f2radiowidget = w;
		range_set_sensitivity((int) client_data);
		if (pr_current_pageradio == 2)
		    XmProcessTraversal(r7text1, XmTRAVERSE_CURRENT);
	    }
	}
	else if (w == f2radiowidget)
	    XmToggleButtonGadgetSetState(w, True, False);
}

/* ARGSUSED */
void
Act_print P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
	char		*ofstring;
	static	Widget	r2radio, r3radio;
	static	Widget	r6radio, r7radio;

	if (print_active) {
	    XRaiseWindow(DISP, XtWindow(print_shell));
	    return;
	}

	if (printlog_active) {
	    XRaiseWindow(DISP, XtWindow(printlog_shell));
	    return;
	}

	if (dvi_file == NULL) {
	    WARN(XmDIALOG_ERROR, "No active dvi file");
	    return;
	}

	ofstring = (pageno_correct == 1 ? mprintf("of %d", total_pages)
	  : mprintf("of %d to %d", pageno_correct,
	    total_pages + pageno_correct - 1));

	if (print_shell == NULL) {
	    Widget	form;
	    Widget	frame1, f1label, f1child;
	    Widget	r4label;
	    Widget	r8label;
	    Widget	frame2, f2label, f2child;
	    Widget	r9ok, r9cancel;
	    XmString	str;

	    XtAddActions(print_actions, XtNumber(print_actions));
	    print_shell = XtVaCreatePopupShell("print",
	      xmDialogShellWidgetClass, top_level,
	      XmNtitle, "Print dvi file",
	      XmNallowShellResize, True,
	      XmNvisual, our_visual,
	      XmNcolormap, our_colormap,
	      NULL);
	    form = XtVaCreateWidget("form", xmFormWidgetClass, print_shell,
	      XmNhorizontalSpacing, DDIST_MAJOR,
	      XmNverticalSpacing, DDIST_MAJOR,
	      XmNautoUnmanage, False,
	      NULL);

	    /*  First frame (rows 1-3):  print to printer or file */

	    frame1 = XtVaCreateWidget("printtoframe", xmFrameWidgetClass, form,
	      XmNmarginWidth, DDIST,
	      XmNmarginHeight, DDIST,
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNrightAttachment, XmATTACH_FORM,
	      NULL);

	    str = XmStringCreateLocalized("Print To");
	    f1label = XtVaCreateManagedWidget("title", xmLabelGadgetClass,
	      frame1,
	      XmNchildType, XmFRAME_TITLE_CHILD,
	      XmNlabelString, str,
	      NULL);
	    XmStringFree(str);

	    f1child = XtVaCreateWidget("form", xmFormWidgetClass, frame1,
	      XmNhorizontalSpacing, DDIST,
	      XmNverticalSpacing, DDIST,
	      NULL);

	    str = XmStringCreateLocalized("Printer");
	    f1radiowidget = r2radio = XtVaCreateManagedWidget("toprinter",
	      xmToggleButtonGadgetClass, f1child,
	      XmNlabelString, str,
	      XmNindicatorType, XmONE_OF_MANY,
	      XmNset, True,
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNtopOffset, 0,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNleftOffset, 0,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r2radio, XmNvalueChangedCallback, cb_print_vs_file,
	      (XtPointer) 1);

	    str = XmStringCreateLocalized("Print command:");
	    r2label = XtVaCreateManagedWidget("prncommand", xmLabelGadgetClass,
	      f1child,
	      XmNlabelString, str,
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNtopOffset, 0,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r2radio,
	      NULL);
	    XmStringFree(str);

	    r2text = XtVaCreateManagedWidget("prntext", xmTextFieldWidgetClass,
	      f1child,
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNtopOffset, 0,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r2label,
	      NULL);
	    XtAddCallback(r2text, XmNactivateCallback, cb_print, NULL);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(r2text, XmNheight, &h2, NULL);
		XtVaGetValues(r2radio, XmNheight, &h1, NULL);
		XtVaSetValues(r2radio, XmNtopOffset, (h2 - h1) / 2, NULL);
		XtVaGetValues(r2label, XmNheight, &h1, NULL);
		XtVaSetValues(r2label, XmNtopOffset, (h2 - h1) / 2, NULL);
	    }

	    str = XmStringCreateLocalized("File");
	    r3radio = XtVaCreateManagedWidget("tofile",
	      xmToggleButtonGadgetClass, f1child,
	      XmNlabelString, str,
	      XmNindicatorType, XmONE_OF_MANY,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r2text,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNleftOffset, 0,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r3radio, XmNvalueChangedCallback, cb_print_vs_file,
	      (XtPointer) 2);

	    str = XmStringCreateLocalized("File name:");
	    r3label = XtVaCreateManagedWidget("filename", xmLabelGadgetClass,
	      f1child,
	      XmNlabelString, str,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r2text,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r3radio,
	      NULL);
	    XmStringFree(str);

	    r3text = XtVaCreateManagedWidget("prntext", xmTextFieldWidgetClass,
	      f1child,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r2text,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r3label,
	      NULL);
	    XtAddCallback(r3text, XmNactivateCallback, cb_print, NULL);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(r3text, XmNheight, &h2, NULL);
		XtVaGetValues(r3radio, XmNheight, &h1, NULL);
		XtVaSetValues(r3radio, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
		XtVaGetValues(r3label, XmNheight, &h1, NULL);
		XtVaSetValues(r3label, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
	    }

	    {	/* align left edges of text widgets */
		Dimension w2r, w3r, w2l, w3l;

		XtVaGetValues(r2radio, XmNwidth, &w2r, NULL);
		XtVaGetValues(r3radio, XmNwidth, &w3r, NULL);
		XtVaGetValues(r2label, XmNwidth, &w2l, NULL);
		XtVaGetValues(r3label, XmNwidth, &w3l, NULL);
		if (w2l < w3l || w2r < w3r) {
		    Dimension offset;

		    offset = DDIST;
		    if (w2l < w3l) offset += w3l - w2l;
		    if (w2r < w3r) offset += w3r - w2r;
		    XtVaSetValues(r2label, XmNleftOffset, offset, NULL);
		}
		if (w2l > w3l || w2r > w3r) {
		    Dimension offset;

		    offset = DDIST;
		    if (w2l > w3l) offset += w2l - w3l;
		    if (w2r > w3r) offset += w2r - w3r;
		    XtVaSetValues(r3label, XmNleftOffset, offset, NULL);
		}
	    }

	    XtManageChild(f1child);
	    XtManageChild(frame1);

	    /* Row 4:  printer name (for dvips) */

	    str = XmStringCreateLocalized("Printer name (used by dvips):");
	    r4label = XtVaCreateManagedWidget("prname", xmLabelGadgetClass,
	      form,
	      XmNlabelString, str,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, frame1,
	      XmNleftAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);

	    r4text = XtVaCreateManagedWidget("prtext", xmTextFieldWidgetClass,
	      form,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, frame1,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r4label,
	      XmNcolumns, 5,
	      NULL);
	    XtAddCallback(r4text, XmNactivateCallback, cb_print, NULL);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(r4text, XmNheight, &h2, NULL);
		XtVaGetValues(r4label, XmNheight, &h1, NULL);
		XtVaSetValues(r4label,
		  XmNtopOffset, DDIST_MAJOR + (h2 - h1) / 2,
		  NULL);
	    }

	    /* Row 8 (row 4.5 for Motif):  additional dvips args */

	    str = XmStringCreateLocalized(
	      "Additional dvips arguments (optional):");
	    r8label = XtVaCreateManagedWidget("xargsname", xmLabelGadgetClass,
	      form,
	      XmNlabelString, str,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r4text,
	      XmNleftAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);

	    r8text = XtVaCreateManagedWidget("xargstext",
	      xmTextFieldWidgetClass, form,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r8label,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNleftOffset, DDIST + DDIST_MAJOR,
	      XmNrightAttachment, XmATTACH_FORM,
	      XmNrightOffset, DDIST + DDIST_MAJOR,
	      NULL);
	    XtAddCallback(r8text, XmNactivateCallback, cb_print, NULL);

	    /* Second frame (rows 5-7):  page selection */

	    frame2 = XtVaCreateWidget("printtoframe", xmFrameWidgetClass, form,
	      XmNmarginWidth, DDIST,
	      XmNmarginHeight, DDIST,
	      XmNresizable, True,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r8text,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNrightAttachment, XmATTACH_FORM,
	      NULL);

	    str = XmStringCreateLocalized("Page Selection");
	    f2label = XtVaCreateManagedWidget("title", xmLabelGadgetClass,
	      frame2,
	      XmNchildType, XmFRAME_TITLE_CHILD,
	      XmNlabelString, str,
	      NULL);
	    XmStringFree(str);

	    f2child = XtVaCreateWidget("form", xmFormWidgetClass, frame2,
	      XmNhorizontalSpacing, DDIST,
	      XmNverticalSpacing, DDIST,
	      XmNresizable, True,
	      NULL);

	    str = XmStringCreateLocalized("All");
	    f2radiowidget = r6radio = XtVaCreateManagedWidget("all",
	      xmToggleButtonGadgetClass, f2child,
	      XmNlabelString, str,
	      XmNindicatorType, XmONE_OF_MANY,
	      XmNset, True,
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNtopOffset, 0,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNleftOffset, 0,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r6radio, XmNvalueChangedCallback, cb_range,
	      (XtPointer) 1);

	    str = XmStringCreateLocalized("Range:");
	    r7radio = XtVaCreateManagedWidget("range",
	      xmToggleButtonGadgetClass, f2child,
	      XmNlabelString, str,
	      XmNindicatorType, XmONE_OF_MANY,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNleftOffset, 0,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r7radio, XmNvalueChangedCallback, cb_range,
	      (XtPointer) 2);

	    str = XmStringCreateLocalized("From");
	    r7label1 = XtVaCreateManagedWidget("from", xmLabelGadgetClass,
	      f2child,
	      XmNlabelString, str,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r7radio,
	      NULL);
	    XmStringFree(str);

	    r7text1 = XtVaCreateManagedWidget("frompage",
	      xmTextFieldWidgetClass, f2child,
	      XmNcolumns, 5,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r7label1,
	      NULL);
	    XtAddCallback(r7text1, XmNactivateCallback, cb_print, NULL);

	    str = XmStringCreateLocalized("to");
	    r7label2 = XtVaCreateManagedWidget("to", xmLabelGadgetClass,
	      f2child,
	      XmNlabelString, str,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r7text1,
	      NULL);
	    XmStringFree(str);

	    r7text2 = XtVaCreateManagedWidget("topage", xmTextFieldWidgetClass,
	      f2child,
	      XmNcolumns, 5,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r7label2,
	      NULL);
	    XtAddCallback(r7text2, XmNactivateCallback, cb_print, NULL);

	    str = XmStringCreateLocalized(ofstring);
	    r7label3 = XtVaCreateManagedWidget("of", xmLabelGadgetClass,
	      f2child,
	      XmNlabelString, str,
	      XmNsensitive, False,
	      XmNresizable, True,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, r6radio,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r7text2,
	      NULL);
	    XmStringFree(str);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(r7text1, XmNheight, &h2, NULL);
		XtVaGetValues(r7radio, XmNheight, &h1, NULL);
		XtVaSetValues(r7radio, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
		XtVaGetValues(r7label1, XmNheight, &h1, NULL);
		XtVaSetValues(r7label1, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
		XtVaSetValues(r7label2, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
		XtVaSetValues(r7label3, XmNtopOffset, DDIST + (h2 - h1) / 2,
		  NULL);
	    }

	    XtManageChild(f2child);
	    XtManageChild(frame2);

	    str = XmStringCreateLocalized("OK");
	    r9ok = XtVaCreateManagedWidget("ok", xmPushButtonWidgetClass,
	      form,
	      XmNlabelString, str,
	      XmNshowAsDefault, True,
	      XmNdefaultButtonShadowThickness, 1,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, frame2,
	      XmNtopOffset, DDIST_MAJOR + DDIST,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNbottomAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r9ok, XmNactivateCallback, cb_print, NULL);
	    XtOverrideTranslations(r9ok, XtParseTranslationTable(
	      "<Key>Return:ArmAndActivate()"));
	    XmProcessTraversal(r9ok, XmTRAVERSE_CURRENT);

	    str = XmStringCreateLocalized("Show previous log");
	    r9prev = XtVaCreateManagedWidget("prev",
	      xmPushButtonWidgetClass, form,
	      XmNlabelString, str,
	      XmNsensitive, False,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, frame2,
	      XmNtopOffset, DDIST_MAJOR + DDIST,
	      XmNleftAttachment, XmATTACH_WIDGET,
	      XmNleftWidget, r9ok,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r9prev, XmNactivateCallback, cb_print_cancel, NULL);

	    str = XmStringCreateLocalized("Cancel");
	    r9cancel = XtVaCreateManagedWidget("cancel",
	      xmPushButtonWidgetClass, form,
	      XmNlabelString, str,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, frame2,
	      XmNtopOffset, DDIST_MAJOR + DDIST,
	      XmNrightAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(r9cancel, XmNactivateCallback, cb_print_cancel, NULL);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(r9ok, XmNheight, &h1, NULL);
		XtVaGetValues(r9prev, XmNheight, &h2, NULL);
		XtVaSetValues(r9prev,
		  XmNtopOffset, DDIST_MAJOR + DDIST + (h1 - h2) / 2, NULL);
		XtVaGetValues(r9cancel, XmNheight, &h2, NULL);
		XtVaSetValues(r9cancel,
		  XmNtopOffset, DDIST_MAJOR + DDIST + (h1 - h2) / 2, NULL);
	    }

	    XtManageChild(form);

	    {	/* center middle button in bottom row */
		Dimension	w, b;
		Position	x1, x3;

		XtVaGetValues(r9ok, XmNx, &x1, XmNwidth, &w, NULL);
		x1 += w;
		XtVaGetValues(r9prev, XmNwidth, &w, NULL);
		XtVaGetValues(r9cancel, XmNx, &x3, NULL);
		XtVaSetValues(r9prev, XmNleftOffset, (x3 - x1 - w) / 2, NULL);
	    }
	}
	else {	/* if the window was already created */
	    XmString	str;

	    if (pr_save_fileradio != pr_current_fileradio)
		XmToggleButtonGadgetSetState(
		  pr_save_fileradio == 1 ? r2radio : r3radio, True, True);
	    if (pr_save_pageradio != pr_current_pageradio)
		XmToggleButtonGadgetSetState(
		  pr_save_pageradio == 1 ? r6radio : r7radio, True, True);

	    str = XmStringCreateLocalized(ofstring);
	    XtVaSetValues(r7label3, XmNlabelString, str, NULL);
	    XmStringFree(str);

	    if (pr_save_printer != NULL) {
		XtVaSetValues(r4text, XmNvalue, pr_save_printer,
		  XmNcursorPosition, strlen(pr_save_printer), NULL);
	    }
	    else
		XtVaSetValues(r4text, XmNvalue, "", NULL);

	    if (pr_save_xargs != NULL) {
		XtVaSetValues(r8text, XmNvalue, pr_save_xargs,
		  XmNcursorPosition, strlen(pr_save_xargs), NULL);
	    }
	    else
		XtVaSetValues(r8text, XmNvalue, "", NULL);
	}

	free(ofstring);

	{	/* set initial values of print command and file name fields */
	    _Xconst char *s;
	    char s2[3 * sizeof(int) + 1];

	    s = pr_save_cmd;
	    if (s == NULL) s = "lpr";
	    XtVaSetValues(r2text, XmNvalue, s, XmNcursorPosition, strlen(s),
	      NULL);

	    if (pr_save_file == NULL) {
		char		*s2;
		int		n;

		s = rindex(dvi_name, '/');
		if (s == NULL) s = dvi_name;
		else ++s;
		n = strlen(s);
		if (n >= 4 && memicmp(s + n - 4, ".dvi", 4) == 0) n -= 4;
		s2 = xmalloc(n + 4);
		memcpy(s2, s, n);
		memcpy(s2 + n, ".ps", 4);
		pr_save_file = s2;
	    }
	    s = pr_save_file;
	    XtVaSetValues(r3text, XmNvalue, s, XmNcursorPosition, strlen(s),
	      NULL);

#if HAVE_VSNPRINTF
	    (void) snprintf(s2, sizeof s2, "%d", current_page + pageno_correct);
#else
	    (void) sprintf(s2, "%d", current_page + pageno_correct);
#endif
	    XtVaSetValues(r7text1, XmNvalue, s2, XmNcursorPosition, strlen(s2),
	      NULL);
	    XtVaSetValues(r7text2, XmNvalue, s2, XmNcursorPosition, strlen(s2),
	      NULL);
	}

	XtPopup(print_shell, XtGrabNone);
	print_active = True;
}

#endif /* MOTIF */


/*
 *	The actual printing
 */

static	int	pageno1, pageno2;		/* page range */
static	Widget	print_confirm = NULL;
static	Widget	printlog_text;
static	Widget	printlog_close;
static	Widget	printlog_cancel;

#if XAW
static	XawTextPosition	printlog_length;
#else /* MOTIF */
static	XmTextPosition	printlog_length;
#endif

static	void		dvips_ended ARGS((int));
static	struct xchild	print_child	= {NULL, 0, True, dvips_ended};

static	void		read_from_dvips ARGS((void));
static	struct xio	print_xio	= {NULL, 0, XIO_IN,
#if HAVE_POLL
					   NULL,
#endif
					   read_from_dvips, NULL};

static	void		dvips_alarm ARGS((void));
static	struct xtimer	dvips_timer	= {NULL, {0, 0}, dvips_alarm};
static	int		dvips_status;
#define	DVIPS_STAT_NONE	0
#define	DVIPS_STAT_RUN	1
#define	DVIPS_STAT_WAIT	2
static	int		dvips_sig;	/* SIGINT or SIGKILL */

/* Forward reference */

static void print_do_it();

static int
getpageno(w)
	Widget	w;
{
	char	*s, *p;

#if XAW
	XtVaGetValues(w, XtNstring, &s, NULL);
#else /* MOTIF */
	s = XmTextFieldGetString(w);
#endif
	p = s;
	if (*p == '-') ++p;
	if (!isdigit(*p)) return 0;
	do ++p; while (isdigit(*p));
	if (*p != '\0') return 0;

#if XAW
	return atoi(s) - pageno_correct + 1;
#else /* MOTIF */
	{
	    int i = atoi(s) - pageno_correct + 1;

	    XtFree(s);
	    return i;
	}
#endif
}

/* ARGSUSED */
static	void
cb_print_confirm(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	XtDestroyWidget(print_confirm);
	print_confirm = NULL;

	if (client_data != NULL)
	    print_do_it();
}

static	void
print_precheck()
{
	char		*dest;
	struct stat	statbuf;

	/* Check for active confirm box */
	if (print_confirm != NULL) {
	    XRaiseWindow(DISP, XtWindow(print_confirm));
	    return;
	}

	/* Validate page range (if given) */
	if (pr_current_pageradio == 2) {	/* if page range selected */
	    pageno1 = getpageno(r7text1);
	    pageno2 = getpageno(r7text2);
	    if (pageno1 <= 0 || pageno1 > pageno2 || pageno2 > total_pages) {
		WARN(XmDIALOG_ERROR, "Invalid page range");
		return;
	    }
	}

	/* Pop down window */
	cb_print_cancel(NULL, NULL, NULL);

	/* Check for whether to clobber existing file */
	if (pr_current_fileradio == 2) {	/* if print to file */
#if XAW
	    XtVaGetValues(r3text, XtNstring, &dest, NULL);
#else /* MOTIF */
	    dest = XmTextFieldGetString(r3text);
#endif
	    if (stat(dest, &statbuf) == 0 && S_ISREG(statbuf.st_mode)) {
		char	*msg;

		msg = mprintf("Overwrite existing file %s?", dest);
		print_confirm = confirm_popup(msg, "OK", cb_print_confirm,
		  (XtPointer) print_shell, NULL);
		free(msg);
#if MOTIF
		XtFree(dest);
#endif
		return;
	    }
#if MOTIF
	    XtFree(dest);
#endif
	}

	print_do_it();
}


static	void	cb_dvips_close ARGS((Widget, XtPointer, XtPointer));
static	void	cb_dvips_keep ARGS((Widget, XtPointer, XtPointer));
static	void	cb_dvips_cancel ARGS((Widget, XtPointer, XtPointer));
static	void	printlog_append ARGS((_Xconst char *, int));

struct dvips_env {
	_Xconst char	*envname;
	_Xconst char	*en1, *en2;
	_Xconst char	*envval;
};

static	struct dvips_env	dvips_envs[] = {
	{NULL, "TEXPICTS", "TEXINPUTS", NULL},
	{NULL, "TEXPSHEADERS", "PSHEADERS", NULL}};

static	void
print_do_it()
{
	_Xconst char	*dest;
	_Xconst char	*prn;
	_Xconst char	*xargs;
	FILE		*f;
	unsigned int	len;	/* length of command line */
	unsigned int	argc;
	char		**argv;
	char		**argnext;
	int		print_io[2];
	char		*p, **pp;
	_Xconst char	*q;

	if (dvi_file == NULL) {
	    WARN(XmDIALOG_ERROR, "No active dvi file");
	    return;
	}

	/* Save state of window for next time */
	pr_save_fileradio = pr_current_fileradio;
	pr_save_pageradio = pr_current_pageradio;

#if XAW

	if (pr_save_fileradio == 1) {	/* if print to printer */
	    if (pr_save_cmd != NULL)
		free((char *) pr_save_cmd);
	    XtVaGetValues(r2text, XtNstring, &dest, NULL);
	    dest = pr_save_cmd = xstrdup(dest);
	}
	else {		/* if print to file */
	    if (pr_save_file != NULL)
		free((char *) pr_save_file);
	    XtVaGetValues(r3text, XtNstring, &dest, NULL);
	    dest = pr_save_file = xstrdup(dest);
	}

	XtVaGetValues(r4text, XtNstring, &prn, NULL);
	if (pr_save_printer != NULL)
	    free((char *) pr_save_printer);
	if (*prn != '\0')
	    prn = pr_save_printer = xstrdup(prn);
	else
	    prn = pr_save_printer = NULL;

	XtVaGetValues(r8text, XtNstring, &xargs, NULL);
	if (pr_save_xargs != NULL)
	    free((char *) pr_save_xargs);
	if (*xargs != '\0')
	    xargs = pr_save_xargs = xstrdup(xargs);
	else
	    xargs = pr_save_xargs = NULL;

#else /* MOTIF */

	if (pr_save_fileradio == 1) {	/* if print to printer */
	    if (pr_save_cmd != NULL)
		XtFree((char *) pr_save_cmd);
	    dest = pr_save_cmd = XmTextFieldGetString(r2text);
	}
	else {		/* if print to file */
	    if (pr_save_file != NULL)
		XtFree((char *) pr_save_file);
	    dest = pr_save_file = XmTextFieldGetString(r3text);
	}

	if (pr_save_printer != NULL)
	    XtFree((char *) pr_save_printer);
	prn = XmTextFieldGetString(r4text);
	if (*prn == '\0') {
	    XtFree((char *) prn);
	    prn = NULL;
	}
	pr_save_printer = prn;

	if (pr_save_xargs != NULL)
	    XtFree((char *) pr_save_xargs);
	xargs = XmTextFieldGetString(r8text);
	if (*xargs == '\0') {
	    XtFree((char *) xargs);
	    xargs = NULL;
	}
	pr_save_xargs = xargs;

#endif

	/* Open file for writing (if necessary) */
	if (pr_save_fileradio == 2) {	/* if print to file */
	    f = xfopen(dest, "w");
	    if (f == NULL) {
		WARN2(XmDIALOG_ERROR, "Error opening file: %s\n%s.", dest,
		  strerror(errno));
		return;
	    }
	}

#if XAW
	/* Create popup window */
	if (printlog_shell == NULL) {
	    Widget		form;
	    int			ddist;
	    Dimension		w0, w1, w2, w3, b;

	    XtSetSensitive(r9prev, True);
	    printlog_shell = XtVaCreatePopupShell("printlog",
	      transientShellWidgetClass, top_level,
	      XtNtitle, "Xdvi print process",
	      XtNmappedWhenManaged, False,
	      XtNtransientFor, top_level,
	      XtNvisual, our_visual,
	      XtNcolormap, our_colormap,
	      NULL);
	    form = XtCreateManagedWidget("form", formWidgetClass,
	      printlog_shell, NULL, 0);
	    printlog_text = XtVaCreateManagedWidget("text",
	      asciiTextWidgetClass, form,
	      XtNstring, "",
	      XtNdataCompression, False,
	      XtNeditType, XawtextAppend,
	      XtNscrollHorizontal, XawtextScrollAlways,
	      XtNscrollVertical, XawtextScrollAlways,
	      XtNwidth, 600,
	      XtNheight, 400,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainRight,
	      XtNtop, XawChainTop,
	      XtNbottom, XawChainBottom,
	      NULL);
	    XtOverrideTranslations(printlog_text, XtParseTranslationTable(
	      "<Key>Return:printlogIntClose()\n\
^<Key>c:printlogIntCancel()\n\
^<Key>s:printlogIntKeep()\n\
^<Key>q:printlogIntUnkeep()"));

	    printlog_close = XtVaCreateManagedWidget("close",
	      commandWidgetClass, form,
	      XtNlabel, "Close window",
	      XtNsensitive, False,
	      XtNfromVert, printlog_text,
	      XtNleft, XawChainLeft,
	      XtNright, XawChainLeft,
	      XtNtop, XawChainBottom,
	      XtNbottom, XawChainBottom,
	      NULL);
	    XtAddCallback(printlog_close, XtNcallback, cb_dvips_close, NULL);

	    printlog_keep = XtVaCreateManagedWidget("keep", toggleWidgetClass,
	      form,
	      XtNlabel, "Keep window open",
	      XtNfromVert, printlog_text,
	      XtNfromHoriz, printlog_close,
	      XtNtop, XawChainBottom,
	      XtNbottom, XawChainBottom,
	      NULL);
	    XtAddCallback(printlog_keep, XtNcallback, cb_dvips_keep, NULL);

	    printlog_cancel = XtVaCreateManagedWidget("cancel",
	      commandWidgetClass, form,
	      XtNlabel, "Cancel",
	      XtNfromVert, printlog_text,
	      XtNfromHoriz, printlog_keep,
	      XtNleft, XawChainRight,
	      XtNright, XawChainRight,
	      XtNtop, XawChainBottom,
	      XtNbottom, XawChainBottom,
	      NULL);
	    XtAddCallback(printlog_cancel, XtNcallback, cb_dvips_cancel, NULL);

	    /* Move the buttons to the right spots */
	    XtVaGetValues(printlog_text, XtNwidth, &w0, NULL);
	    XtVaGetValues(form, XtNdefaultDistance, &ddist, NULL);
	    XtVaGetValues(printlog_close, XtNwidth, &w1, XtNborderWidth, &b,
	      NULL);
	    w1 += 2 * b + ddist;
	    XtVaGetValues(printlog_keep, XtNwidth, &w2, XtNborderWidth, &b,
	      NULL);
	    w2 += 2 * b;
	    if (w0 > w2 + 2 * w1) {
		XtVaSetValues(printlog_keep,
		  XtNhorizDistance, (w0 - w2) / 2 - w1, NULL);
		w1 += (w0 - w2) / 2 - w1 - ddist;
	    }
	    w1 += w2;
	    XtVaGetValues(printlog_cancel, XtNwidth, &w3, XtNborderWidth, &b,
	      NULL);
	    w3 += 2 * b;
	    if (w1 + ddist + w3 < w0)
		XtVaSetValues(printlog_cancel, XtNhorizDistance, w0 - w1 - w3,
		  NULL);
	}
	else {
	    XtVaSetValues(printlog_text, XtNstring, "", NULL);
	    XtSetSensitive(printlog_close, False);
	    XtSetSensitive(printlog_keep, True);
	    XtSetSensitive(printlog_cancel, True);
	}

#else /* MOTIF */

	/* Create popup window */
	if (printlog_shell == NULL) {
	    Widget	form;
	    XmString	str;

	    XtSetSensitive(r9prev, True);
	    printlog_shell = XtVaCreatePopupShell("printlog",
	      xmDialogShellWidgetClass, top_level,
	      XmNtitle, "Xdvi print process",
	      XmNvisual, our_visual,
	      XmNcolormap, our_colormap,
	      NULL);
	    form = XtVaCreateWidget("form", xmFormWidgetClass, printlog_shell,
	      XmNhorizontalSpacing, DDIST,
	      XmNverticalSpacing, DDIST,
	      XmNautoUnmanage, False,
	      NULL);

	    printlog_text = XmCreateScrolledText(form, "text", NULL, 0);
	    XtVaSetValues(XtParent(printlog_text),
	      XmNtopAttachment, XmATTACH_FORM,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNrightAttachment, XmATTACH_FORM,
	      NULL);
	    XtVaSetValues(printlog_text,
	      XmNeditable, False,
	      XmNrows, 24,
	      XmNcolumns, 80,
	      XmNeditMode, XmMULTI_LINE_EDIT,
	      NULL);
	    XtOverrideTranslations(printlog_text, XtParseTranslationTable(
	      "<Key>Return:printlogIntClose()\n\
^<Key>c:printlogIntCancel()\n\
^<Key>s:printlogIntKeep()\n\
^<Key>q:printlogIntUnkeep()"));
	    XtManageChild(printlog_text);

	    str = XmStringCreateLocalized(
	      "Keep window open after dvips finishes");
	    printlog_keep = XtVaCreateManagedWidget("keep",
	      xmToggleButtonGadgetClass, form,
	      XmNlabelString, str,
	      XmNnavigationType, XmTAB_GROUP,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, XtParent(printlog_text),
	      XmNleftAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(printlog_keep, XmNvalueChangedCallback, cb_dvips_keep,
	      NULL);

	    str = XmStringCreateLocalized("Close window");
	    printlog_close = XtVaCreateManagedWidget("close",
	      xmPushButtonWidgetClass, form,
	      XmNlabelString, str,
	      XmNshowAsDefault, True,
	      XmNsensitive, False,
	      XmNdefaultButtonShadowThickness, 1,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, printlog_keep,
	      XmNtopOffset, DDIST,
	      XmNleftAttachment, XmATTACH_FORM,
	      XmNbottomAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(printlog_close, XmNactivateCallback, cb_dvips_close,
	      NULL);

	    str = XmStringCreateLocalized("Cancel");
	    printlog_cancel = XtVaCreateManagedWidget("cancel",
	      xmPushButtonWidgetClass, form,
	      XmNlabelString, str,
	      XmNtopAttachment, XmATTACH_WIDGET,
	      XmNtopWidget, printlog_keep,
	      XmNtopOffset, DDIST,
	      XmNrightAttachment, XmATTACH_FORM,
	      NULL);
	    XmStringFree(str);
	    XtAddCallback(printlog_cancel, XmNactivateCallback, cb_dvips_cancel,
	      NULL);

	    {	/* straighten the row of widgets */
		Dimension h1, h2;

		XtVaGetValues(printlog_close, XmNheight, &h1, NULL);
		XtVaGetValues(printlog_cancel, XmNheight, &h2, NULL);
		XtVaSetValues(printlog_cancel,
		  XmNtopOffset, DDIST + (h1 - h2) / 2, NULL);
		XtVaSetValues(printlog_cancel,
		  XmNrightOffset, DDIST + (h1 - h2) / 2, NULL);
	    }

	    XtManageChild(form);
	    XmProcessTraversal(printlog_text, XmTRAVERSE_CURRENT);
	}
	else {
	    XmTextSetString(printlog_text, "");
	    XtSetSensitive(printlog_close, False);
	    XtSetSensitive(printlog_keep, True);
	    XtSetSensitive(printlog_cancel, True);
	}

#endif

	/* Compute length of dvips command line */
	/*   dvips [-Pps] -f [-o!command] [-p=n -lm] [xargs] */
	argc = 3;
	len = strlen(resource.dvips_path) + 4;
	  /* "dvips -f\0" */
	if (prn != NULL) argc = 4, len += strlen(prn) + 3;
	if (pr_save_fileradio == 1) ++argc, len += strlen(dest) + 4;
	if (pr_save_pageradio == 2) {
	    argc += 2;
	    len += uintlen(pageno1) + uintlen(pageno2) + 7;
	}
	if (pr_save_xargs != NULL) {
	    _Xconst char *q = pr_save_xargs;

	    len += strlen(pr_save_xargs);
	    for (;;) {
		while (*q == ' ' || *q == '\t') ++q, --len;
		if (*q == '\0') break;
		++argc;
		++len;
		while (*q != '\0' && *q != ' ' && *q != '\t') ++q;
	    }
	}

	/* Compute actual dvips command line */
	argv = xmalloc(argc * sizeof(char *));
	p = argv[0] = xmalloc(len + 1);
	argnext = argv + 1;
	p += strlen(strcpy(p, resource.dvips_path));
	*p++ = ' ';
	*argnext++ = p;

	if (prn != NULL) {
	    sprintf(p, "-P%s ", prn);
	    p += strlen(p);
	    *argnext++ = p;
	}

	*p++ = '-';
	*p++ = 'f';
	*p++ = ' ';
	*argnext++ = p;

	if (pr_save_fileradio == 1) {
	    sprintf(p, "-o!%s ", dest);
	    p += strlen(p);
	    *argnext++ = p;
	}

	if (pr_save_pageradio == 2) {
	    sprintf(p, "-p=%u ", pageno1);
	    p += strlen(p);
	    *argnext++ = p;
	    sprintf(p, "-l%u ", pageno2);
	    p += strlen(p);
	    *argnext++ = p;
	}

	if (pr_save_xargs != NULL) {
	    _Xconst char *q = pr_save_xargs;

	    for (;;) {
		while (*q == ' ' || *q == '\t') ++q;
		if (*q == '\0') break;
		while (*q != '\0' && *q != ' ' && *q != '\t') *p++ = *q++;
		*p++ = ' ';
		*argnext++ = p;
	    }
	}

#if 0
	if (argnext - argv != argc)
	    oops("dvips command count mismatch:  %u != %u", argc,
	      argnext - argv);
	if (p - argv[0] != len)
	    oops("dvips command length mismatch:  %u != %u", len, p - argv[0]);
#endif

	p[-1] = '\n';
#if MOTIF
	*p = '\0';
#endif

	printlog_length = 0;

	q = rindex(dvi_name, '/');
	if (q != NULL) {
	    struct dvips_env *dep;

#if PS
	    /* grab copies of environment variables before we change them */
	    ps_init_paths();
#endif

	    if (dvips_envs[0].envname == NULL) {
		for (dep = dvips_envs; dep < dvips_envs + XtNumber(dvips_envs);
		  ++dep) {
		    _Xconst char *envname;
		    _Xconst char *envval;
		    unsigned int len;

		    envval = getenv(envname = dep->en1);
		    if (envval == NULL) {
			envval = getenv(dep->en2);
			if (envval != NULL) envname = dep->en2;
		    }
		    dep->envname = envname;
		    len = q - dvi_name + 2;
		    if (envval != NULL)
			len += strlen(envval);
		    dep->envval = p = xmalloc(len);
		    bcopy(dvi_name, p, q - dvi_name);
		    p += q - dvi_name;
		    *p++ = ':';
		    *p = '\0';
		    if (envval != NULL)
			Strcpy(p, envval);
		    xputenv(dep->envname, dep->envval);
		}
	    }
	    for (dep = dvips_envs; dep < dvips_envs + XtNumber(dvips_envs);
	      ++dep) {
		printlog_append(dep->envname, strlen(dep->envname));
		printlog_append("=", 1);
		printlog_append(dep->envval, strlen(dep->envval));
		printlog_append(" ", 1);
	    }
	}

	printlog_append(argv[0], len);
#if XAW
	do_popup(printlog_shell);
#else
	XtPopup(printlog_shell, XtGrabNone);
#endif
	printlog_active = True;

	--argnext;
	for (pp = argv + 1; pp <= argnext; ++pp) (*pp)[-1] = '\0';
	*argnext = NULL;

	if (xpipe(print_io) != 0) {
	    perror("[xdvi] pipe");
	    return;
	}

	/* Fork process */
	Fflush(stderr);		/* avoid double buffering */
	print_child.pid = vfork();
	if (print_child.pid == 0) {		/* if child */
	    (void) close(0);
	    (void) dup2(fileno(dvi_file), 0);
	    (void) close(fileno(dvi_file));
	    (void) lseek(0, 0, SEEK_SET);
	    (void) close(1);
	    if (pr_save_fileradio == 1) {	/* if printing to printer */
		(void) dup2(print_io[1], 1);
	    }
	    else {	/* printing to file */
		(void) dup2(fileno(f), 1);
		(void) close(fileno(f));
	    }
	    (void) close(2);
	    (void) dup2(print_io[1], 2);
	    (void) close(print_io[1]);
	    (void) close(print_io[0]);

	    if (setsid() == -1) {	/* so we can kill the process group */
		perror("setsid");
		Fflush(stderr);
		_exit(1);
	    }
	    (void) execvp(*argv, argv);
	    Fprintf(stderr, "Execvp of %s failed.\n", *argv);
	    Fflush(stderr);
	    _exit(1);
	}

	free(argv[0]);
	free(argv);
	if (pr_save_fileradio == 2) {
	    fclose(f);
	    ++n_files_left;
	}

	if (print_child.pid == -1) {	/* error */
	    perror("[xdvi] vfork");
	    return;
	}

	set_chld(&print_child);
	dvips_sig = SIGINT;

	(void) close(print_io[1]);
	++n_files_left;

	/* Set up file descriptor for non-blocking I/O */
	prep_fd(print_io[0], True);
	print_xio.fd = print_io[0];
	set_io(&print_xio);

	dvips_status = DVIPS_STAT_RUN;	/* running */
}


#if XAW

static void
printlog_append(str, len)
	_Xconst char	*str;
	int		len;
{
	static	XawTextBlock	block	= {0, 0, NULL, 0};

	block.ptr = (char *) str;
	block.length = len;
	block.format = XawFmt8Bit;
	while (XawTextReplace(printlog_text, printlog_length, printlog_length,
	  &block) != XawEditDone) {
	    int length;

	    XtVaGetValues(printlog_text, XtNlength, &length, NULL);
	    printlog_length = length;
	}
	printlog_length += len;
	XawTextSetInsertionPoint(printlog_text, printlog_length);
}

#else /* MOTIF */

static void
printlog_append(str, len)
	_Xconst char	*str;
	int		len;
{
	XmTextInsert(printlog_text, printlog_length, (char *) str);
	printlog_length += len;
	XtVaSetValues(printlog_text, XmNcursorPosition, printlog_length, NULL);
	XmTextShowPosition(printlog_text, printlog_length);
}

#endif

static void
read_from_dvips()
{
	int	bytes;
	char	line[80];

	for (;;) {
#if XAW
	    bytes = read(print_xio.fd, line, sizeof line);
#else
	    bytes = read(print_xio.fd, line, sizeof line - 1);
#endif
	    if (bytes < 0) {
		if (AGAIN_CONDITION)
		    break;
		perror("xdvi: read_from_dvips");
		break;
	    }

	    if (bytes == 0)
		break;
	    else {
#if MOTIF
		line[bytes] = '\0';
#endif
		printlog_append(line, bytes);
	    }
	}
}

/* ARGSUSED */
static	void
cb_dvips_cancel(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	if (dvips_status != DVIPS_STAT_RUN)
	    return;	/* How did we get here? */

	kill(print_child.pid, dvips_sig);
	dvips_sig = SIGKILL;
	printlog_append("^C", 2);
}

static void
dvips_ended(status)
{
	char	*str;
	int	ms;

	read_from_dvips();
	clear_io(&print_xio);
	(void) close(print_xio.fd);
	++n_files_left;

	if (WIFEXITED(status)) {
	    if (WEXITSTATUS(status) == 0) {
		printlog_append("Done.\n", 6);
		str = NULL;
	    }
	    else
		str = mprintf("\nPrint process returned exit code %d.\n",
		  WEXITSTATUS(status));
	}
	else if (WIFSIGNALED(status))
	    str = mprintf("\nPrint process terminated by signal %d.\n",
	      WTERMSIG(status));
	else
	    str = mprintf("\nPrint process returned unknown status 0x%x.\n",
	      status);

	ms = resource.dvips_hang;
	if (str != NULL) {
	    ms = resource.dvips_fail_hang;
	    printlog_append(str, strlen(str));
	    free(str);
	}

	if (ms > 0) {
	    set_timer(&dvips_timer, ms);
	    dvips_status = DVIPS_STAT_WAIT;
	}
	else {
	    printlog_act_keep(NULL, NULL, NULL, NULL);
	    dvips_status = DVIPS_STAT_NONE;
	}

	XtSetSensitive(printlog_close, True);
	XtSetSensitive(printlog_cancel, False);
}

static void
dvips_pop_down()
{
	XtPopdown(printlog_shell);
	printlog_active = False;
}

static void
dvips_alarm()
{
#if XAW
	Boolean	state;
#endif

#if XAW
	XtVaGetValues(printlog_keep, XtNstate, &state, NULL);
	if (!state)
	    dvips_pop_down();
#else /* MOTIF */
	if (!XmToggleButtonGadgetGetState(printlog_keep))
	    dvips_pop_down();
#endif
	dvips_status = DVIPS_STAT_NONE;
}

/* ARGSUSED */
static	void
cb_dvips_close(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
	if (dvips_status == DVIPS_STAT_RUN)
	    return;	/* How did we get here? */

	if (dvips_status == DVIPS_STAT_WAIT) {
	    dvips_status = DVIPS_STAT_NONE;
	    cancel_timer(&dvips_timer);
	}

	dvips_pop_down();
}

/* ARGSUSED */
static	void
cb_dvips_keep(w, client_data, call_data)
	Widget		w;
	XtPointer	client_data;
	XtPointer	call_data;
{
#if XAW
	Boolean	state;
#endif

	if (dvips_status != DVIPS_STAT_NONE)
	    return;	/* Nothing to do */

#if XAW
	XtVaGetValues(printlog_keep, XtNstate, &state, NULL);
	if (!state)
	    dvips_pop_down();
#else
	if (!XmToggleButtonGadgetGetState(printlog_keep))
	    dvips_pop_down();
#endif
}

/* ARGSUSED */
static	void
printlog_act_close P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
	cb_dvips_close(NULL, NULL, NULL);
}

/* ARGSUSED */
static	void
printlog_act_keep P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
#if XAW
	XtVaSetValues(printlog_keep, XtNstate, True, NULL);
#else /* MOTIF */
	XmToggleButtonGadgetSetState(printlog_keep, True, False);
#endif
}

/* ARGSUSED */
static	void
printlog_act_unkeep P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
#if XAW
	XtVaSetValues(printlog_keep, XtNstate, False, NULL);
#else /* MOTIF */
	XmToggleButtonGadgetSetState(printlog_keep, False, False);
#endif
	if (dvips_status == DVIPS_STAT_NONE)
	    dvips_pop_down();
}

/* ARGSUSED */
static	void
printlog_act_cancel P4C(Widget, w, XEvent *, event, String *, params,
  Cardinal *, num_params)
{
	cb_dvips_cancel(NULL, NULL, NULL);
}
